/*******************************************************************************
* File............: bbc2txt.c                                                  *
* Version.........: $Id: 2c4dd6185852650cc780ac5a7b568b7d3173f298 $            *
* Created.........: 2019-06-22                                                 *
* Last changed....: 2019-06-23                                                 *
* Author..........: TheBlackzone <theblackzone@gmail.com>                      *
*------------------------------------------------------------------------------*
* Purpose:                                                                     *
*   Converts tokenized Acorn BBC BASIC V files to text                         *
*------------------------------------------------------------------------------*
* Compile:                                                                     *
*   gcc -o bbc2txt bbc2txt.c                                                   *
*------------------------------------------------------------------------------*
* Usage:                                                                       *
*   bbc2txt [basicv file]                                                      *
*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>


/*******************************************************************************
* 1 byte tokens                                                                *
*******************************************************************************/

static const char *t1[] = {
    "OTHERWISE",    /* 0x7F */
    "AND",          /* 0x80 */
    "DIV",          /* 0x81 */
    "EOR",          /* 0x82 */
    "MOD",          /* 0x83 */
    "OR",           /* 0x84 */
    "ERROR",        /* 0x85 */
    "LINE",         /* 0x86 */
    "OFF",          /* 0x87 */
    "STEP",         /* 0x88 */
    "SPC",          /* 0x89 */
    "TAB(",         /* 0x8A */
    "ELSE",         /* 0x8B */
    "THEN",         /* 0x8C */
    "INTRNL",       /* 0x8D - used in internal-format line numbers (not a token) */
    "OPENIN",       /* 0x8E */
    "PTR",          /* 0x8F */
    "PAGE",         /* 0x90 */
    "TIME",         /* 0x91 */
    "LOMEM",        /* 0x92 */
    "HIMEM",        /* 0x93 */
    "ABS",          /* 0x94 */
    "ACS",          /* 0x95 */
    "ADVAL",        /* 0x96 */
    "ASC",          /* 0x97 */
    "ASN",          /* 0x98 */
    "ATN",          /* 0x99 */
    "BGET",         /* 0x9A */
    "COS",          /* 0x9B */
    "COUNT",        /* 0x9C */
    "DEG",          /* 0x9D */
    "ERL",          /* 0x9E */
    "ERR",          /* 0x9F */
    "EVAL",         /* 0xA0 */
    "EXP",          /* 0xA1 */
    "EXT",          /* 0xA2 */
    "FALSE",        /* 0xA3 */
    "FN",           /* 0xA4 */
    "GET",          /* 0xA5 */
    "INKEY",        /* 0xA6 */
    "INSTR(",       /* 0xA7 */
    "INT",          /* 0xA8 */
    "LEN",          /* 0xA9 */
    "LN",           /* 0xAA */
    "LOG",          /* 0xAB */
    "NOT",          /* 0xAC */
    "OPENUP",       /* 0xAD */
    "OPENOUT",      /* 0xAE */
    "PI",           /* 0xAF */
    "POINT(",       /* 0xB0 */
    "POS",          /* 0xB1 */
    "RAD",          /* 0xB2 */
    "RND",          /* 0xB3 */
    "SGN",          /* 0xB4 */
    "SIN",          /* 0xB5 */
    "SQR",          /* 0xB6 */
    "TAN",          /* 0xB7 */
    "TO",           /* 0xB8 */
    "TRUE",         /* 0xB9 */
    "USR",          /* 0XBA */
    "VAL",          /* 0xBB */
    "VPOS",         /* 0xBC */
    "CHR$",         /* 0xBD */
    "GET$",         /* 0xBE */       
    "INKEY$",       /* 0xBF */
    "LEFT$(",       /* 0xC0 */
    "MID$(",        /* 0xC1 */
    "RIGHT$(",      /* 0xC2 */
    "STR$",         /* 0xC3 */
    "STRING$(",     /* 0xC4 */
    "EOF",          /* 0xC5 */
    "ESCFUNC",      /* 0xC6 - escape code for 2-byte functions (not a token) */
    "ESCCMND",      /* 0xC7 - escape code for 2-byte commands (not a token) */
    "ESCSTMT",      /* 0xC8 - escape code for 2-byte statements (not a token) */
    "WHEN",         /* 0xC9 */
    "OF",           /* 0xCA */
    "ENDCASE",      /* 0xCB */
    "ELSE",         /* 0xCC */
    "ENDIF",        /* 0xCD */
    "ENDWHILE",     /* 0xCE */
    "PTR#",         /* 0xCF */
    "PAGE",         /* 0xD0 */
    "TIME",         /* 0xD1 */
    "LOMMEM",       /* 0xD2 */
    "HIMMEM",       /* 0xD3 */
    "SOUND",        /* 0xD4 */
    "BPUT",         /* 0xD5 */
    "CALL",         /* 0xD6 */
    "CHAIN",        /* 0xD7 */
    "CLEAR",        /* 0xD8 */
    "CLOSE",        /* 0xD9 */
    "CLG",          /* 0xDA */
    "CLS",          /* 0xDB */
    "DATA",         /* 0xDC */
    "DEF",          /* 0xDD */
    "DIM",          /* 0xDE */
    "DRAW",         /* 0xDF */
    "END",          /* 0xE0 */
    "ENDPROC",      /* 0xE1 */
    "ENVELOPE",     /* 0xE2 */
    "FOR",          /* 0xE3 */
    "GOSUB",        /* 0xE4 */
    "GOTO",         /* 0xE5 */
    "GCOL",         /* 0xE6 */
    "IF",           /* 0xE7 */
    "INPUT",        /* 0xE8 */
    "LET",          /* 0xE9 */
    "LOCAL",        /* 0xEA */
    "MODE",         /* 0xEB */
    "MOVE",         /* 0xEC */
    "NEXT",         /* 0xED */
    "ON",           /* 0xEE */
    "VDU",          /* 0xEF */
    "PLOT",         /* 0xF0 */
    "PRINT",        /* 0xF1 */
    "PROC",         /* 0xF2 */
    "READ",         /* 0xF3 */
    "REM",          /* 0xF4 */
    "REPEAT",       /* 0xF5 */
    "REPORT",       /* 0xF6 */
    "RESTORE",      /* 0xF7 */
    "RETURN",       /* 0xF8 */
    "RUN",          /* 0xF9 */
    "STOP",         /* 0xFA */
    "COLOUR",       /* 0xFB */
    "TRACE",        /* 0xFC */
    "UNTIL",        /* 0xFD */
    "WIDTH",        /* 0xFE */
    "OSCLI"         /* 0xFF */
};


/*******************************************************************************
* 2 byte function tokens, following the 0xC6 ESCFN byte                        *
*******************************************************************************/

static const char *t2func[] = {
    "SUM",          /* 0x8E */
    "BEAT"          /* 0x8F */
};


/*******************************************************************************
* 2 byte command tokens, following the 0xC7 ESCCOM byte                        *
*******************************************************************************/

static const char * t2cmd[] = {
    "APPEND",       /* 0x8E */
    "AUTO",         /* 0x8F */
    "CRUNCH",       /* 0x90 */
    "DELETE",       /* 0x91 */
    "EDIT",         /* 0x92 */
    "HELP",         /* 0x93 */
    "LIST",         /* 0x94 */
    "LOAD",         /* 0x95 */
    "LVAR",         /* 0x96 */
    "NEW",          /* 0x97 */
    "OLD",          /* 0x98 */
    "RENUMBER",     /* 0x99 */
    "SAVE",         /* 0x9A */
    "TEXTLOAD",     /* 0x9B */
    "TEXTSAVE",     /* 0x9C */
    "TWIN",         /* 0x9D */
    "TWINO",        /* 0x9E */
    "INSTALL"       /* 0x9F */
};


/*******************************************************************************
* 2 byte statement tokens, following the 0xC8 ESCSTMT byte                     *
*******************************************************************************/

static const char *t2stmt[] = {
    "CASE",         /* 0x8E */
    "CIRCLE",       /* 0x8F */
    "FILL",         /* 0x90 */
    "ORIGIN",       /* 0x91 */
    "POINT",        /* 0x92 */
    "RECTANGLE",    /* 0x93 */
    "SWAP",         /* 0x94 */
    "WHILE",        /* 0x95 */
    "WAIT",         /* 0x96 */
    "MOUSE",        /* 0x97 */
    "QUIT",         /* 0x98 */
    "SYS",          /* 0x99 */
    "INSTALL",      /* 0x9A */
    "LIBRARY",      /* 0x9B */
    "TINT",         /* 0x9C */
    "ELLIPSE",      /* 0x9D */
    "BEATS",        /* 0x9E */
    "TEMPO",        /* 0x9F */
    "VOICES",       /* 0xA0 */
    "VOICE",        /* 0xA1 */
    "STEREO",       /* 0xA2 */
    "OVERLAY"       /* 0xA3 */
};


/*******************************************************************************
* main                                                                         *
*******************************************************************************/

int main(int argc, char **argv)
{
	FILE *f;
	int ch;
	int i;
	int hi,lo,lineno,len;
	int b0,b1,b2,lineref;

	if (argc > 1) {
		if(strcmp(argv[1],"-") == 0) {
			f = stdin;
		} else {
			f = fopen(argv[1], "r");
				if (f == NULL) {
					fprintf(stderr, "Cannot open file '%s': %s\n", argv[1], strerror(errno));
					exit(EXIT_FAILURE);
				}
		}
	} else {
		f = stdin;
	}

	i = 0;
	len = 0;
	while(feof(f) == 0) {

		ch = fgetc(f);

		/* check the first byte if it matches the 0x0D "start line" token */
		if ((i == 0) && (ch != 0x0D)) {
			fprintf(stderr, "This doesn't look like a BBC BASIC file.\n");
			if (f != stdin) {
				(void)fclose(f);
			}
			exit(EXIT_FAILURE);
		}

		if (ch == 0x0d) { /* start of line */
			hi = fgetc(f);

			/* 0x0D,0xFF marks the end of a BASIC file */
			if (hi == 0xFF)
				break;

			lo = fgetc(f);
			lineno = hi * 256 + lo;
			printf("%5d ", lineno);
			len = fgetc(f) - 4;
			i += 4;
		} else if (ch == 0x8D) { /* line reference */
			b0 = fgetc(f);
			b1 = fgetc(f);
			b2 = fgetc(f);
			lineref = ((b2 ^ (b0 * 16)) * 256 + (b1 ^ ((b0 * 4) & 0xC0))) & 0xFFFF;
			printf("%d", lineref);
			len -= 4;
			i += 4;
		} else if (ch == 0xC6) { /* 2 byte function token */
			ch = fgetc(f);
			printf("%s", t2func[ch - 0x8E]);
			len -= 2;
			i += 2;
		} else if (ch == 0xC7) { /* 2 byte command  token */
			ch = fgetc(f);
			printf("%s", t2cmd[ch - 0x8E]);
			len -= 2;
			i += 2;
		} else if (ch == 0xC8) { /* 2 byte statement token */
			ch = fgetc(f);
			printf("%s", t2stmt[ch - 0x8E]);
			len -= 2;
			i += 2;
		} else if (ch >= 0x7F) { /* 1 byte token */
			printf("%s", t1[ch - 0x7F]);
			len--;
			i++;
		} else {
			printf("%c", ch);
			len--;
			i++;
		}
		if (len == 0) {
			printf("\n");
		}
	}

	if (f != stdin) {
		(void)fclose(f);
	}

    exit(EXIT_SUCCESS);

} /* main */

